/*
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 *    MSSRegression.java
 *    Copyright (C) 2017 University of Patras, Greece
 *
 */

package weka.classifiers.meta;

import weka.classifiers.AbstractClassifier;
import weka.classifiers.Classifier;
import weka.classifiers.ParallelMultipleClassifiersCombiner;
import weka.classifiers.trees.M5P;
import weka.core.Capabilities;
import weka.core.CommandlineRunnable;
import weka.core.Environment;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.RevisionUtils;
import weka.core.SelectedTag;
import weka.core.Tag;
import weka.core.TechnicalInformation;
import weka.core.Capabilities.Capability;
import weka.core.TechnicalInformation.Field;
import weka.core.TechnicalInformation.Type;
import weka.core.TechnicalInformationHandler;
import weka.core.Utils;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Vector;

/**
 * <!-- globalinfo-start --> Class implementing multi-scheme semi-supervised regression.<br/>
 * <br/>
 * For more information see:<br/>
 * <br/>
 * N. Fazakis, S. Karlos, S. Kotsiantis & K. Sgarbas.
 * A multi-scheme semi-supervised regression approach.
 * (2017) (Submitted)
 * <p/>
 * <!-- globalinfo-end -->
 * 
 * <!-- options-start --> Valid options are:
 * <p/>
 *
 * <pre>
 * -B &lt;classifier specification&gt;
 *  Full class name of base classifier to include, followed
 *  by scheme options. May be specified multiple times.
 *  (default: -B "weka.classifiers.trees.M5P" -B "weka.classifiers.rules.M5P"
 *  -B "weka.classifiers.functions.SMOreg" -B "weka.classifiers.trees.RandomForest")
 * </pre>
 * 
 * <pre>
 * -do-not-check-capabilities
 *  If set, classifier capabilities are not checked before classifier is built
 *  (use with caution).
 * </pre>
 * 
 * <pre>
 * -R &lt;RAN&gt;
 *  The combination rule to use
 *  (default: RAN)
 * </pre>
 * 
 * <pre>
 * -M &lt;num&gt;
 *  The maximum number of iterations allowed to perform in selftrain
 *  (default: 10)
 * </pre>
 * 
 * <pre>
 * -L &lt;num&gt;
 *  The labeled ratio
 *  (default: 0.25)
 * </pre>
 *
 * <pre>
 * -do-not-print
 *  If set, no individual models are printed in the output
 * </pre> 
 * 
 * <pre>
 * -W &lt;classifier specification&gt;
 *  Full class name of the final classifier to use, followed
 *  by scheme options.
 *  (default: "weka.classifiers.trees.M5P")
 * </pre>
 * 
 * <!-- options-end -->
 * 
 * <!-- technical-bibtex-start --> BibTeX:
 * 
 * <pre>
 * &#64;article{Fazakis2017,
 *    author = {N. Fazakis, S. Karlos, S. Kotsiantis & K. Sgarbas.},
 *    journal = {(Submitted)},
 *    number = {X},
 *    pages = {XX-XX},
 *    title = {A multi-scheme semi-supervised regression approach},
 *    volume = {XX},
 *    year = {2017}
 * }
 * </pre>
 * <p/>
 * <!-- technical-bibtex-end -->
 * 
 * @author Nikos Fazakis (fazakis@ece.upatras.gr)
 * @author Stamatis Karlos (stkarlos@upatras.gr)
 * @author Sotiris Kotsiantis (sotos@math.upatras.gr)
 * @author Kyriakos Sgarbas (sgrabas@upatras.gr)	
 * @version $Revision: 10000 $
 */
public class MSSRegression extends ParallelMultipleClassifiersCombiner implements
  TechnicalInformationHandler {

  /** For serialization */
  private static final long serialVersionUID = 3291055894654301318L;

  /** Combination rule: Minimum Range */
  public static final int MINIMUM_RANGE = 1;
  
  public static final Tag[] TAGS_RULES = {
    new Tag(MINIMUM_RANGE, "RAN", "Minimum Range"),
     };

  /** Combination Rule variable */
  //TODO: support multiple combination rules
  protected int m_CombinationRule = MINIMUM_RANGE;
    
  /** Environment variables */
  protected transient Environment m_env = Environment.getSystemWide();

  /** Structure of the training data */
  protected Instances m_structure;

  /** Print the individual models in the output */
  protected boolean m_dontPrintModels;
  
  /** Number of maximum iterations*/
  protected int m_maxIter=10;
  
  /** Labeled ratio*/
  protected double m_percentLabeled=0.25;
  
  /**
   * Constructor.
   */
  public MSSRegression() {
	  Classifier[] defaultClassifiers = {
			    new weka.classifiers.trees.M5P(),
			    new weka.classifiers.functions.SMOreg(),
			    new weka.classifiers.trees.RandomForest()
			  };
	  setClassifiers(defaultClassifiers);
  }

  /**
   * Returns a string describing MSSRegression
   * 
   * @return a description suitable for displaying in the explorer/experimenter
   *         gui
   */
  public String globalInfo() {
    return "Class implementing multi-scheme semi-supervised regression."
      + "\n\n"
      + "For more information see:\n\n" + getTechnicalInformation().toString();
  }

  /**
   * Returns an enumeration describing the available options.
   * 
   * @return an enumeration of all the available options.
   */
  @Override
  public Enumeration<Option> listOptions() {

    Vector<Option> result = new Vector<Option>();
    
    /** These options are inherited**/
    result.addElement(new Option(
            "\tNumber of execution slots.\n"
            + "\t(default 1 - i.e. no parallelism)",
            "num-slots", 1, "-num-slots <num>"));
    
    result.addElement(new Option(
            "\tFull class name of classifier to include, followed\n"
            + "\tby scheme options. May be specified multiple times.\n"
            + "\t(default: -B \"weka.classifiers.rules.M5P\""
            +" -B \"weka.classifiers.functions.SMOreg\" -B \"weka.classifiers.trees.RandomForest\")",
            "B", 1, "-B <classifier specification>"));    
    
    final class Temp extends AbstractClassifier{
    	@Override
		public void buildClassifier(Instances data) throws Exception {}};
    result.addAll(Collections.list((new Temp()).listOptions()));
    /** End inherited options **/
    
    result.addElement(new Option("\tThe combination rule to use\n"
      + "\t(default: RAN)", "R", 1, "-R " + Tag.toOptionList(TAGS_RULES)));    
    
    result.addElement(new Option("\tThe maximum number of iterations allowed to perform in selftrain.\n"
    	      + "\t(default: 10)", "M", 1, "-M <num>"));
    
    result.addElement(new Option("\tThe labeled ratio.\n"
  	      + "\t(default: 0.25)", "L", 1, "-L <num>"));
    
    result.addElement(new Option(
    	      "\tSuppress the printing of the individual models in the output",
    	      "do-not-print", 1, "-do-not-print"));
    
    result.addElement(new Option(
            "\tFull name of final classifier.\n"
            + "\t(default: " + finalClassifierString() + 
            ((finalClassifierOptions().length > 0) ? 
             " with options " + Utils.joinOptions(finalClassifierOptions()) + ")" : ")"),
            "W", 1, "-W"));

    /** It gets too crowded with this super**/
    //result.addAll(Collections.list(super.listOptions()));

    return result.elements();
  }

  /**
   * Gets the current settings of MSSRegression.
   * 
   * @return an array of strings suitable for passing to setOptions()
   */
  @Override
  public String[] getOptions() {
    int i;
    Vector<String> result = new Vector<String>();
    String[] options;

    options = super.getOptions();
    for (i = 0; i < options.length; i++) {
      result.add(options[i]);
    }

    result.add("-R");
    result.add("" + getCombinationRule());

    result.add("-M");
    result.add("" + getMaxIter());
    
    result.add("-L");
    result.add("" + getPercentLabeled());    

    if (m_dontPrintModels) {
      result.add("-do-not-print");
    }
    
    /* This must be last to work correctly! */
    result.add("-W");
    result.add(getFinalClassifier().getClass().getName());
    
    String[] finalClassifierOptions = ((OptionHandler)f_Classifier).getOptions();
    if (finalClassifierOptions.length > 0) {
    	result.add("--");
      Collections.addAll(result, finalClassifierOptions);
    }

    return result.toArray(new String[result.size()]);
  }

  /**
   * Parses a given list of options.
   * <p/>
   * 
   * <!-- options-start --> Valid options are:
   * <p/>
   * 
   * <pre>
   * -B &lt;classifier specification&gt;
   *  Full class name of classifier to include, followed
   *  by scheme options. May be specified multiple times.
   *  (default: -B "weka.classifiers.trees.M5P" -B "weka.classifiers.rules.M5P"
   *  -B "weka.classifiers.functions.SMOreg" -B "weka.classifiers.trees.RandomForest")
   * </pre>
   * 
   * <pre>
   * -do-not-check-capabilities
   *  If set, classifier capabilities are not checked before classifier is built
   *  (use with caution).
   * </pre>
   * 
   * <pre>
   * -R &lt;RAN&gt;
   *  The combination rule to use
   *  (default: RAN)
   * </pre>
   * 
   * <pre>
   * -M &lt;num&gt;
   *  The maximum number of iterations allowed to perform in selftrain
   *  (default: 10)
   * </pre>
   * 
   * <pre>
   * -L &lt;num&gt;
   *  The labeled ratio
   *  (default: 0.25)
   * </pre>
   *
   * <pre>
   * -do-not-print
   *  If set, no individual models are printed in the output
   * </pre>
   * 
   * <pre>
   * -W &lt;classifier specification&gt;
   *  Full class name of the final classifier to use, followed
   *  by scheme options.
   *  (default: "weka.classifiers.trees.M5P")
   * </pre>
   * 
   * <!-- options-end -->
   * 
   * @param options the list of options as an array of strings
   * @throws Exception if an option is not supported
   */
  @Override
  public void setOptions(String[] options) throws Exception {
	  
	String finalClassifierName = Utils.getOption('W', options);

	if (finalClassifierName.length() > 0) {
	  setFinalClassifier(AbstractClassifier.forName(finalClassifierName, null));
	  setFinalClassifier(AbstractClassifier.forName(finalClassifierName,
	        Utils.partitionOptions(options)));
	} else {
	  setFinalClassifier(AbstractClassifier.forName(finalClassifierString(), null));
	  String[] fClassifierOptions = Utils.partitionOptions(options);
	  if (fClassifierOptions.length > 0) {
	   setFinalClassifier(AbstractClassifier.forName(finalClassifierString(),
	                                                 fClassifierOptions));
	  } else {
	     setFinalClassifier(AbstractClassifier.forName(finalClassifierString(),
	                                                 finalClassifierOptions()));
	  }
	}		  
	  
    String tmpStr;

    tmpStr = Utils.getOption('R', options);
    if (tmpStr.length() != 0) {
      setCombinationRule(new SelectedTag(tmpStr, TAGS_RULES));
    } else {
      setCombinationRule(new SelectedTag(MINIMUM_RANGE, TAGS_RULES));
    }
    
    String maxIter = Utils.getOption('M', options);
    if (maxIter.length() != 0) {
      setMaxIter(Integer.parseInt(maxIter));
    } else {
    	setMaxIter(10);
    }
    
    String percentLabeled = Utils.getOption('L', options);
    if (percentLabeled.length() != 0) {
    	setPercentLabeled(Double.parseDouble(percentLabeled));
    } else {
    	setPercentLabeled(0.25);
    }

    setDoNotPrintModels(Utils.getFlag("do-not-print", options));

    super.setOptions(options);
  }

  /**
   * Returns an instance of a TechnicalInformation object, containing detailed
   * information about the technical background of this class, e.g., paper
   * reference or book this class is based on.
   * 
   * @return the technical information about this class
   */
  @Override
  public TechnicalInformation getTechnicalInformation() {
    TechnicalInformation result;

    result = new TechnicalInformation(Type.ARTICLE);
    result.setValue(Field.AUTHOR, "N. Fazakis, S. Karlos, S. Kotsiantis & K. Sgarbas");
    result.setValue(Field.TITLE,
      "A multi-scheme semi-supervised regression approach");
    result.setValue(Field.JOURNAL,
      "Submitted");
    result.setValue(Field.VOLUME, "XX");
    result.setValue(Field.NUMBER, "X");
    result.setValue(Field.PAGES, "XX-XX");
    result.setValue(Field.YEAR, "2017");
    result.setValue(Field.PUBLISHER, "Submitted");

    return result;
  }

  /**
   * Returns capabilities according to the selected classifiers.
   * 
   * @return the capabilities of this classifier
   */
  @Override
  public Capabilities getCapabilities() {
	Capabilities result = super.getCapabilities();
    result.disableDependency(Capability.NOMINAL_CLASS);
    result.disable(Capability.NOMINAL_CLASS);

    return result;
  }

  /**
   * Buildclassifier trains the classifiers iteratively and builds the final classifier.
   * 
   * @param data the training data to be used
   * @throws Exception if the classifier could not be built successfully
   */
  @Override
  public void buildClassifier(Instances data) throws Exception {
	  
	//TODO: support multiple base classifiers  
	if(m_Classifiers.length!=3)
		throw new Exception("Number of base classifiers should be 3 explicitly.");
			
	double percentLabeled = m_percentLabeled;
	double maxPercentUnlabeledFromLabeled=0.1;
	double percentBestUnlabeled = maxPercentUnlabeledFromLabeled*percentLabeled;
	int maxIter = m_maxIter;
	
	double percentUnlabeled = 1-percentLabeled;
	int numLabeled = new Double(percentLabeled*data.numInstances()).intValue();
	int numUnlabeled = new Double(percentUnlabeled*data.numInstances()).intValue();	
	int numOfBestUnlabeled = new Double(percentBestUnlabeled*data.numInstances()).intValue();
	
	Instances labeled = new Instances(data,0, numLabeled);
	labeled.deleteWithMissingClass();
	Instances unlabeled = new Instances(data,numLabeled,numUnlabeled);
	
	for(int i=0;i<maxIter && unlabeled.numInstances()>0 ;i++){		
		//build classifiers sequentially
		/*for (int j = 0; j < m_Classifiers.length; j++) {
			getClassifier(j).buildClassifier(labeled);
		}*/		
		//build classifiers in parallel
		super.buildClassifier(labeled);
		try{
			super.buildClassifiers(labeled);
		}catch(Exception e){
			if(e.getMessage().toLowerCase().contains("all class values are the same"))				
				throw new Exception(e.getMessage()+"\nTry increasing labeled ratio!");
			else
				throw e;
		}
		
		//save predictions
		Double[][] pred = new Double[unlabeled.numInstances()][3];
		Double[][] predAv = new Double[unlabeled.numInstances()][3];
		//Double[] predWeights = new Double[3];
		for(int j=0;j<unlabeled.numInstances();j++){			
			pred[j][0] = getClassifier(0).classifyInstance(unlabeled.instance(j));
			pred[j][1] = getClassifier(1).classifyInstance(unlabeled.instance(j));
			pred[j][2] = getClassifier(2).classifyInstance(unlabeled.instance(j));
			
			double max =Math.max(Math.max(pred[j][0], pred[j][1]),pred[j][2]);
			double min =Math.min(Math.min(pred[j][0], pred[j][1]),pred[j][2]);
			predAv[j][0]=Math.abs(max-min); //range
			predAv[j][1]=(double) j; //position
			predAv[j][2]=(pred[j][0]+pred[j][1]+pred[j][2])/3.0; //average prediction
			
			//optimize
			pred[j]=null;
		}
		//System.out.println("Calculated..");
		
		//pick the best unlabeled and add them to labeled
		Arrays.sort(predAv,new Comparator<Double[]>(){
			@Override
			public int compare(final Double[] entry1,final Double[] entry2){
				final Double range1 = entry1[0];
				final Double range2 = entry2[0];
				return range1.compareTo(range2);
			}
		});
		
		// Add best unlabeled to labeled and after remove them (there's a reason for 2 steps)
		int maxBestUnlabeled = Math.min(predAv.length, numOfBestUnlabeled); // unlabeled must be enough
		Integer[] pointerDeleteUnlabeled = new Integer[maxBestUnlabeled];
		//System.out.println("Added: "+maxBestUnlabeled);
		for(int k=0;k<maxBestUnlabeled;k++){
			//System.out.println("aa"+Arrays.deepToString(predAv[k]));						
			unlabeled.instance(new Double(predAv[k][1]).intValue()).setClassValue(predAv[k][2]);
			labeled.add(unlabeled.instance(new Double(predAv[k][1]).intValue()));						
			pointerDeleteUnlabeled[k]=new Double(predAv[k][1]).intValue();
		}			
		// it's important to iterate from last to first, because when we remove
		// an instance, the rest shifts by one position.
		Arrays.sort(pointerDeleteUnlabeled);
		//System.out.println(Arrays.deepToString(pointerDeleteUnlabeled));
		for(int k=maxBestUnlabeled-1;k>=0;k--)
			unlabeled.delete(pointerDeleteUnlabeled[k]);
		
	}
	
	getFinalClassifier().buildClassifier(labeled);	  
	
  }
  
  /** Not used **/
  public static Double[] medianToWeights(Double[] median) {
	    if ( (median[0] - median[1]) * (median[2] - median[0]) >= 0 ) // a >= b and a <= c OR a <= b and a >= c
	        return new Double[]{0.7,0.15,0.15};
	    else if ( (median[1] - median[0]) * (median[2] - median[1]) >= 0 ) // b >= a and b <= c OR b <= a and b >= c
	        return new Double[]{0.15,0.7,0.15};
	    else
	        return new Double[]{0.15,0.15,0.7};
	}

  /**
   * Classifies the given test instance.
   * 
   * @param instance the instance to be classified
   * @return the predicted most likely class for the instance or
   *         Utils.missingValue() if no prediction is made
   * @throws Exception if an error occurred during the prediction
   */
  @Override
  public double classifyInstance(Instance instance) throws Exception {
	  
	  
	 return getFinalClassifier().classifyInstance(instance);
	
  }

  /**
   * Returns the tip text for this property
   * 
   * @return tip text for this property suitable for displaying in the
   *         explorer/experimenter gui
   */
  public String combinationRuleTipText() {
    return "The combination rule used.";
  }

  /**
   * Gets the combination rule used
   * 
   * @return the combination rule used
   */
  public SelectedTag getCombinationRule() {
    return new SelectedTag(m_CombinationRule, TAGS_RULES);
  }

  /**
   * Sets the combination rule to use. Values other than
   * 
   * @param newRule the combination rule method to use
   */
  public void setCombinationRule(SelectedTag newRule) {
    if (newRule.getTags() == TAGS_RULES) {
      m_CombinationRule = newRule.getSelectedTag().getID();
    }
  }

  

  /**
   * Returns the tip text for this property
   * 
   * @return tip text for this property suitable for displaying in the
   *         explorer/experimenter gui
   */
  public String doNotPrintModelsTipText() {
    return "Do not print the individual trees in the output";
  }

  /**
   * Set whether to print the individual ensemble models in the output
   * 
   * @param print true if the individual models are to be printed
   */
  public void setDoNotPrintModels(boolean print) {
    m_dontPrintModels = print;
  }

  /**
   * Get whether to print the individual ensemble models in the output
   * 
   * @return true if the individual models are to be printed
   */
  public boolean getDoNotPrintModels() {
    return m_dontPrintModels;
  }

  /**
   * Output a representation of this classifier
   * 
   * @return a string representation of the classifier
   */
  @Override
  public String toString() {

    if (m_Classifiers == null) {
      return "MSSR: No model built yet.";
    }

    String result = "MSSRegression combines";
    result += " the prediction range of these base learners:\n";
    for (int i = 0; i < m_Classifiers.length; i++) {
      result += '\t' + getClassifierSpec(i) + '\n';
    }

    
    result += "using the '";

    switch (m_CombinationRule) {
    case MINIMUM_RANGE:
      result += "Minimum Range";
      break;

    default:
      throw new IllegalStateException("Unknown combination rule '"
        + m_CombinationRule + "'!");
    }

    result += "' combination rule \n";

    StringBuilder resultBuilder = null;
    if (!m_dontPrintModels) {      
      resultBuilder = new StringBuilder();
      resultBuilder.append(result).append("\nAll the models:\n\n");
      for (Classifier c : m_Classifiers) {
        resultBuilder.append(c).append("\n");
      }
      //resultBuilder.append(result).append("\nNot yet supported\n\n");

    }

    return resultBuilder == null ? result : resultBuilder.toString();
  }

  /**
   * Returns the revision string.
   * 
   * @return the revision
   */
  @Override
  public String getRevision() {
    return RevisionUtils.extract("$Revision: 10000 $");
  }

  /**
   * Main method for testing this class.
   * 
   * @param argv should contain the following arguments: -t training file [-T
   *          test file] [-c class index]
   */
  public static void main(String[] argv) {
    runClassifier(new MSSRegression(), argv);
  }
  
  
  /** MSSR specific options **/

  /**
   * Returns the tip text for this property
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui
   */
  public String percentLabeledTipText() {
    return "The labeled ratio.";
  }
  
  /**
   * Set the labeled ratio.
   *
   * @param the labeled ratio.
   */
  public void setPercentLabeled(double percentLabeled) {

	  m_percentLabeled = percentLabeled;
  }

  /**
   * Get the labeled ratio.
   *
   * @return the labeled ratio.
   */
  public double getPercentLabeled() {

    return m_percentLabeled;
  }
  
  /**
   * Returns the tip text for this property
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui
   */
  public String maxIterTipText() {
    return "The maximum number of iterations allowed to perform in selftrain.";
  }
  
  /**
   * Set maximum iterations.
   *
   * @param the number of maximum iterations.
   */
  public void setMaxIter(int maxIter) {

	  m_maxIter = maxIter;
  }

  /**
   * Get maximum iterations.
   *
   * @return the number of maximum iterations.
   */
  public int getMaxIter() {

    return m_maxIter;
  }
  
  
  /** Final classifier additional functions **/
  
  /** The default final classifier to use */
  protected Classifier f_Classifier = new M5P();

  /**
   * String describing default final classifier.
   */
  protected String finalClassifierString() {

    return "weka.classifiers.trees.M5P";
  }

  /**
   * String describing options for final classifier.
   */
  protected String[] finalClassifierOptions() {

    return new String[0];
  }
  
  /**
   * Returns the tip text for this property
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui
   */
  public String finalClassifierTipText() {
	  
    return "The final classifier to be used for prediction.";
  }
  
  /**
   * Set the final classifier.
   *
   * @param newClassifier the classifier to use.
   */
  public void setFinalClassifier(Classifier newClassifier) {

	  f_Classifier = newClassifier;
  }

  /**
   * Get the classifier used as the final classifier.
   *
   * @return the classifier used as the classifier
   */
  public Classifier getFinalClassifier() {

    return f_Classifier;
  }
  
  /**
   * Gets the final classifier specification string, which contains the class name of
   * the classifier and any options to the classifier
   *
   * @return the classifier string
   */
  protected String getFinalClassifierSpec() {
    Classifier c = getFinalClassifier();
    return c.getClass().getName() + " "
      + Utils.joinOptions(((OptionHandler)c).getOptions());
  }

  @Override
  public void preExecution() throws Exception {
    if (getFinalClassifier() instanceof CommandlineRunnable) {
      ((CommandlineRunnable) getFinalClassifier()).preExecution();
    }
    super.preExecution();
  }

  @Override
  public void postExecution() throws Exception {
    if (getFinalClassifier() instanceof CommandlineRunnable) {
      ((CommandlineRunnable) getFinalClassifier()).postExecution();
    }
    super.postExecution();
  }

}
